import os
import pytest

from . import connect
from .exceptions import ErrorMessageException
from .model import InformationResponse, FetchResponse


def get_connection_string():
    return os.environ.get("CONNECTION_STRING")


def no_connection_string():
    return get_connection_string() is None


@pytest.fixture
async def connection():
    async with connect(get_connection_string()) as c:
        yield c


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
async def test_info(connection):
    info = await connection.info()
    assert isinstance(info, InformationResponse)


actor_id = None


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
@pytest.mark.dependency()
async def test_create_actor(connection):
    info = await connection.info()
    base_url = info.base_urls[0]
    actor_count = len(info.actors)

    assert base_url.startswith("http://") or base_url.startswith("https://")

    actor = await connection.create_actor(base_url)

    assert actor["id"].startswith(base_url)

    global actor_id
    actor_id = actor["id"]

    new_info = await connection.info()
    assert len(new_info.actors) == actor_count + 1
    assert actor_id in new_info.actors


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
@pytest.mark.dependency(depends=["test_create_actor"])
async def test_fetch(connection):
    fetch = await connection.fetch(actor_id, actor_id)
    assert isinstance(fetch, FetchResponse)

    assert fetch.uri == actor_id
    assert fetch.data["id"] == actor_id


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
@pytest.mark.dependency(depends=["test_create_actor"])
async def test_trigger_send_message(connection):
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": actor_id + "#12343",
        "type": "AnimalSound",
        "to": "https://www.w3.org/ns/activitystreams#Public",
        "actor": actor_id,
        "content": "moo",
    }

    await connection.trigger("send_message", {"actor": actor_id, "data": data})

    outgoing = await connection.next_outgoing()

    assert outgoing["actor"] == actor_id
    raw_data = outgoing["data"]["raw"]
    assert raw_data["type"] == "AnimalSound"


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
async def test_error_queue(connection):
    with pytest.raises(ErrorMessageException):
        await connection.create_actor("http://unknown.example")
    await connection.next_error()


@pytest.mark.skipif(no_connection_string(), reason="No connection string")
@pytest.mark.dependency(depends=["test_fetch", "test_trigger_send_message"])
async def test_delete_actor(connection):

    await connection.trigger("delete_actor", {"actor": actor_id})

    outgoing = await connection.next_outgoing()

    raw_data = outgoing["data"]["raw"]
    assert raw_data["type"] == "Delete"

    # FIXME
    #
    # info = await connection.info()
    # assert actor_id not in info.actors



import logging
import json
import ssl

from uuid import uuid4
from typing import List

from contextlib import asynccontextmanager
from dataclasses import dataclass

logger = logging.getLogger(__name__)

try:
    from aiomqtt import Client

    from paho.mqtt.properties import Properties, PacketTypes
    from paho.mqtt.enums import MQTTProtocolVersion

    @dataclass
    class MqttBaseConnection:
        host: str
        port: int

        username: str
        password: str
        path: str
        protocol: str

        echo: bool = False
        client: Client | None = None

        connected: bool = False

        @property
        def transport(self):
            if self.protocol in ["mqtt", "mqtts"]:
                return "tcp"
            else:
                return "websockets"

        @asynccontextmanager
        async def run(self, topics_to_subscribe: List[str]):
            if self.protocol in ["wss", "mqtts"]:
                tls_context = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS)
            else:
                tls_context = None

            self.client = Client(
                self.host,
                port=self.port,
                transport=self.transport,
                username=self.username,
                password=self.password,
                websocket_path=self.path,
                keepalive=50,
                tls_context=tls_context,
                protocol=MQTTProtocolVersion.MQTTv5,
            )

            try:
                async with self.client:
                    self.connected = True
                    for topic in topics_to_subscribe:
                        await self.client.subscribe(topic)
                    logger.info("Connected to %s as %s", self.host, self.username)

                    yield self.client

            except Exception as e:
                logger.exception(e)
                yield None

        async def send(
            self, topic: str, data: dict, correlation_data: str | None = None
        ) -> bytes:
            if correlation_data is None:
                correlation_data = str(uuid4())
            properties = Properties(PacketTypes.PUBLISH)
            properties.CorrelationData = correlation_data.encode()

            if self.echo:
                logger.info("%s: %s, %s", topic, correlation_data, json.dumps(data))

            await self.client.publish(topic, json.dumps(data), properties=properties)

            return correlation_data

except ImportError:

    class MqttBaseConnection:
        @staticmethod
        def from_connection_string(connection_string: str, echo: bool):
            raise NotImplementedError("please install aiomqtt")

        def __init__(self, *args, **kwargs):
            raise NotImplementedError("please install aiomqtt")



import json
import logging

from almabtrieb.util import parse_connection_string
from almabtrieb.base import BaseConnection, ReceivedMessage, MessageType

from .base import MqttBaseConnection

logger = logging.getLogger(__name__)


class MqttConnection(BaseConnection):

    def __init__(
        self,
        base_connection: MqttBaseConnection,
        username: str,
        echo: bool = False,
        silent: bool = False,
    ):
        super().__init__(echo=echo, silent=silent)
        self.username = username
        self.base_connection = base_connection

    @staticmethod
    def from_connection_string(
        connection_string: str, echo: bool = False
    ) -> "MqttConnection":
        parsed = parse_connection_string(connection_string)
        return MqttConnection(
            base_connection=MqttBaseConnection(**parsed, echo=echo),
            echo=echo,
            username=parsed["username"],
        )

    @property
    def connected(self):
        return self.base_connection.connected

    @property
    def receive(self):
        return f"receive/{self.username}/#"

    @property
    def receive_error(self):
        return f"error/{self.username}"

    @property
    def send_topic(self):
        return f"send/{self.username}"

    async def run(self):
        async with self.base_connection.run(
            [self.receive, self.receive_error]
        ) as client:
            async for message in client.messages:
                correlation_id = getattr(message.properties, "CorrelationData").decode()

                if message.topic.matches("receive/+/response/+"):
                    message_type = MessageType.response
                elif message.topic.matches("receive/+/incoming"):
                    message_type = MessageType.incoming
                elif message.topic.matches("receive/+/outgoing"):
                    message_type = MessageType.outgoing
                elif message.topic.matches("error/+"):
                    message_type = MessageType.error
                else:
                    logger.warning("Unknown topic: %s", message.topic)
                    message_type = MessageType.unknown

                await self.handle(
                    ReceivedMessage(
                        message_type=message_type,
                        correlation_id=correlation_id,
                        data=json.loads(message.payload),
                    ),
                )

    async def send(
        self, topic_end: str, data: dict, correlation_data: str | None = None
    ) -> bytes:
        topic = f"{self.send_topic}/{topic_end}"

        correlation_data = await self.base_connection.send(
            topic, data, correlation_data=correlation_data
        )

        return correlation_data



import pytest

from unittest.mock import MagicMock

from .amqp import AmqpConnection, response_regex, error_regex


def is_aio_pika_installed():
    try:
        import aio_pika  # noqa

        return True
    except ImportError:
        return False


@pytest.mark.skipif(not is_aio_pika_installed(), reason="aiopoka not installed")
def test_connected():
    connection = AmqpConnection(MagicMock(), "name")

    assert not connection.connected


def test_response_regex():
    assert response_regex.match("receive.alice.response.fetch")

    assert not response_regex.match("receive.alice.incoming")
    assert not response_regex.match("receive.alice.incoming.test")

    assert error_regex.match("error.alice")



from uuid import uuid4
import json
import logging
import re

from almabtrieb.base import BaseConnection, ReceivedMessage, MessageType


logger = logging.getLogger(__name__)

response_regex = re.compile(r"^receive.\w+\.response\.\w+$")
incoming_regex = re.compile(r"^receive.\w+\.incoming$")
outgoing_regex = re.compile(r"^receive.\w+\.outgoing$")
error_regex = re.compile(r"^error.\w+$")

try:
    import aio_pika

    class AmqpConnection(BaseConnection):
        def __init__(
            self,
            connection_string: str,
            username: str,
            echo: bool = False,
            silent: bool = False,
        ):
            super().__init__(echo=echo, silent=silent)

            self.connection_string = connection_string
            self.channel = None
            self.username = username

        @property
        def connected(self):
            return self.channel is not None

        @property
        def subscription_topic(self):
            return f"receive.{self.username}.#"

        @property
        def subscription_topic_error(self):
            return f"error.{self.username}"

        def routing_key(self, topic_end: str):
            """
            ```pycon
            >>> connection = AmqpConnection("amqp://guest:guest@localhost/", "alice")
            >>> connection.routing_key("example.one")
            'send.alice.example.one'

            >>> connection.routing_key("example/two")
            'send.alice.example.two'

            ```
            """

            return f"send.{self.username}.{topic_end.replace('/', '.')}"

        async def run(self):
            if not self.silent:
                logger.info("Conneting to amqp with %s", self.connection_string)
            connection = await aio_pika.connect_robust(self.connection_string)

            async with connection:
                self.channel = await connection.channel()
                await self.channel.set_qos(prefetch_count=1)
                self.exchange = await self.channel.declare_exchange(
                    "amq.topic",
                    aio_pika.ExchangeType.TOPIC,
                    durable=True,
                )

                queue = await self.channel.declare_queue(
                    "almabtrieb_queue_" + str(uuid4()),
                    durable=False,
                )

                await queue.bind(self.exchange, routing_key=self.subscription_topic)
                await queue.bind(
                    self.exchange, routing_key=self.subscription_topic_error
                )

                async with queue.iterator() as iterator:
                    async for message in iterator:
                        async with message.process():
                            await self.handle_message(message)

        async def handle_message(self, message):
            routing_key = message.routing_key

            if response_regex.match(routing_key):
                message_type = MessageType.response
            elif incoming_regex.match(routing_key):
                message_type = MessageType.incoming
            elif outgoing_regex.match(routing_key):
                message_type = MessageType.outgoing
            elif error_regex.match(routing_key):
                message_type = MessageType.error
            else:
                logger.info("Unknown routing key %s", routing_key)
                message_type = MessageType.unknown

            await self.handle(
                ReceivedMessage(
                    message_type=message_type,
                    correlation_id=message.correlation_id,
                    data=json.loads(message.body),
                ),
            )

        async def send(
            self, topic_end: str, data: dict, correlation_data: str | None = None
        ):
            await self.exchange.publish(
                aio_pika.Message(
                    body=json.dumps(data).encode(),
                    correlation_id=correlation_data,
                ),
                routing_key=self.routing_key(topic_end),
            )

            if self.echo:
                logger.info(
                    "Sent %s %s",
                    self.routing_key(topic_end),
                    json.dumps(data, indent=2),
                )

except ImportError:

    class AmqpConnection:
        def __init__(self, *args, **kwargs):
            raise ImportError("aio_pika is not installed")



import pytest

from . import MqttConnection


def is_mqtt_installed():
    try:
        import aiomqtt  # noqa

        return True
    except ImportError:
        return False


@pytest.mark.skipif(not is_mqtt_installed(), reason="aiomqtt not installed")
def test_from_connection_string():
    connection = MqttConnection.from_connection_string("ws://user:pass@host/ws")

    assert isinstance(connection, MqttConnection)


@pytest.mark.skipif(not is_mqtt_installed(), reason="aiomqtt not installed")
def test_from_connection_string_echo():
    connection = MqttConnection.from_connection_string(
        "ws://user:pass@host/ws", echo=True
    )

    assert isinstance(connection, MqttConnection)



"""
Data models used in communication with the Cattle Drive protocol

"""

from typing import List, Dict, Any
from pydantic import BaseModel, Field, ConfigDict


class InformationResponse(BaseModel):
    """Response for the information request"""

    actors: List[str] = Field(
        examples=[
            [
                "http://host.example/actor/1",
                "http://host.example/actor/2",
            ]
        ],
        description="""Actors of the account on the server""",
    )

    base_urls: List[str] = Field(
        examples=[["http://host.example"]],
        alias="baseUrls",
        description="""The base urls of the server""",
    )


class CreateActorRequest(BaseModel):
    """Request to create an actor"""

    base_url: str = Field(
        examples=["http://host.example"],
        serialization_alias="baseUrl",
        description="""Base url for the actor, the actor URI will be of the form `{base_url}/actor/{id}`""",
    )

    preferred_username: str | None = Field(
        None,
        examples=["alice", "bob"],
        description="""
    Add a preferred username. This name will be used in acct:username@domain and supplied to webfinger. Here domain is determine from baseUrl.
    """,
        serialization_alias="preferredUsername",
    )
    profile: Dict[str, Any] = Field(
        {},
        examples=[{"summary": "A new actor"}],
        description="""
    New profile object for the actor.
    """,
    )
    automatically_accept_followers: bool | None = Field(
        examples=[True],
        description="""
    Enables setting actors to automatically accept follow requests
    """,
        serialization_alias="automaticallyAcceptFollowers",
    )


class WithActor(BaseModel):
    """Used as base for messages requiring an actor"""

    actor: str = Field(
        examples=["http://host.example/actor/1"],
        description="""The actor performing the action""",
    )


class FetchMessage(WithActor):
    """Message to fetch an object from the Fediverse"""

    uri: str = Field(
        examples=["http://remote.example/object/1"],
        description="""The resource to fetch""",
    )


class FetchResponse(WithActor):
    """Result of a a fetch request"""

    uri: str = Field(
        examples=["http://remote.example/object/1"],
        description="""The resource that was requested""",
    )

    data: dict | None = Field(description="""The data returned for the object""")


class TriggerMessage(WithActor):
    """Message to trigger something on the ActivityExchange"""

    model_config = ConfigDict(
        extra="allow",
    )



import asyncio
import logging

from dataclasses import dataclass
from contextlib import asynccontextmanager

from .amqp import AmqpConnection
from .mqtt import MqttConnection
from .model import (
    InformationResponse,
    CreateActorRequest,
    FetchMessage,
    FetchResponse,
)
from .exceptions import NoIncomingException
from .util import ConnectionParams


logger = logging.getLogger(__name__)


@dataclass
class Almabtrieb:
    """Implements the asynchronous API of the Cattle Drive Protocol"""

    connection: MqttConnection | AmqpConnection

    @property
    def connected(self):
        """Is one connected to the service"""
        return self.connection.connected

    @staticmethod
    def from_connection_string(
        connection_string: str, echo: bool = False, silent: bool = False
    ):
        """Creates instance for connection string

        ```python
        Almabtrieb.from_connection_string("ws://user:pass@host/ws")
        Almabtrieb.from_connection_string("amqp://user:pass@host/ws")
        ```

        :param connection_string: The connection string
        :param echo: Set to true to log all messages
        """

        params = ConnectionParams.from_string(connection_string)

        if params.protocol in ["ws", "wss", "mqtt"]:
            return Almabtrieb(
                connection=MqttConnection.from_connection_string(
                    connection_string, echo=echo, silent=silent
                )
            )
        elif params.protocol == "amqp":
            return Almabtrieb(
                connection=AmqpConnection(
                    connection_string=connection_string,
                    echo=echo,
                    username=params.username,
                    silent=silent,
                )
            )
        else:
            raise NotImplementedError("Protocol not implemented")

    async def run(self):
        """Starts the connection and queries for new received messages

        ```python
        task = asyncio.create_task(almabtrieb.run())
        ...
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        ```
        """
        await self.connection.run()

    async def info(self) -> InformationResponse:
        """Returns the information about the connected
        Cattle Drive server

        :returns: The information about the server
        """
        result = await self.connection.send_with_reply("request/info", {})

        return InformationResponse.model_validate(result)

    async def create_actor(
        self,
        base_url: str,
        preferred_username: str | None = None,
        profile: dict = {},
        automatically_accept_followers: bool | None = None,
    ) -> dict:
        """Creates a new actor

        :param base_url: The base url of the actor
        :param preferred_username: The preferred username used as `acct:preferred_username@domain` where `domain` is from `base_url`
        :param profile: The profile of the actor
        :param automatically_accept_followers: If true, the server will automatically accept followers
        :returns: The created actor profile
        """
        request = CreateActorRequest(
            base_url=base_url,
            preferred_username=preferred_username,
            profile=profile,
            automatically_accept_followers=automatically_accept_followers,
        )

        return await self.connection.send_with_reply(
            "request/create_actor", request.model_dump()
        )

    async def fetch(self, actor: str, uri: str, timeout: float = 1) -> FetchResponse:
        """Fetches the object with uri `uri` as the actor with actor id `actor`.

        :param actor: The actor id must be part of actors from the info response
        :param uri: The uri of the object to fetch
        :param timeout: The timeout for the request
        :returns: The fetched object
        """
        async with asyncio.timeout(timeout):
            result = await self.connection.send_with_reply(
                "request/fetch", FetchMessage(actor=actor, uri=uri).model_dump()
            )

            return FetchResponse.model_validate(result)

    async def trigger(self, method: str, data: dict):
        end = "trigger"
        if method and len(method) > 0:
            end = f"{end}/{method}"
        return await self.connection.send(
            end,
            data,
        )

    async def incoming(self):
        """Generator for the incoming messages"""
        while msg := await self.connection.incoming_queue.get():
            yield msg

    async def outgoing(self):
        """Generator for the outgoing messages"""
        while msg := await self.connection.outgoing_queue.get():
            yield msg

    async def next_incoming(self, timeout: float = 1) -> dict:
        """Returns the next incoming message

        :param timeout: The timeout for the request
        :returns: The next incoming message
        """
        try:
            async with asyncio.timeout(timeout):
                return await self.connection.incoming_queue.get()
        except asyncio.TimeoutError:
            raise NoIncomingException("No incoming message")

    async def next_outgoing(self, timeout: float = 1) -> dict:
        """Returns the next outgoing message

        :param timeout: The timeout for the request
        :returns: The next outgoing message
        """
        try:
            async with asyncio.timeout(timeout):
                return await self.connection.outgoing_queue.get()
        except asyncio.TimeoutError:
            raise NoIncomingException("No outgoing message")

    async def next_error(self, timeout: float = 1) -> dict:
        """Returns the next error message

        :param timeout: The timeout for the request
        :returns: The next error message
        """
        try:
            async with asyncio.timeout(timeout):
                return await self.connection.error_queue.get()
        except asyncio.TimeoutError:
            raise NoIncomingException("No error message")

    async def clear_incoming(self) -> None:
        """Empties the incoming message queue"""
        while not self.connection.incoming_queue.empty():
            await self.connection.incoming_queue.get()


@asynccontextmanager
async def connect(connection_string: str, echo: bool = False) -> Almabtrieb:
    """Connects to the service

    Usage
    ```python
    async with connect("ws://user:pass@host/ws") as almabtrieb:
        print(await almabtrieb.info())
    ```
    """
    almabtrieb = Almabtrieb.from_connection_string(
        connection_string=connection_string,
        echo=echo,
    )

    task = asyncio.create_task(almabtrieb.run())
    while not almabtrieb.connected:
        logger.info("Waiting for connection")
        await asyncio.sleep(0.1)
    try:
        yield almabtrieb
    finally:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass



from urllib.parse import urlparse
from dataclasses import dataclass


@dataclass
class ConnectionParams:
    protocol: str
    host: str
    port: int
    username: str | None = None
    password: str | None = None
    path: str | None = None

    @staticmethod
    def from_string(connection_string: str):
        """
        ```pycon
        >>> ConnectionParams.from_string("ws://user:pass@host/ws")
        ConnectionParams(protocol='ws', host='host', port=80, username='user', password='pass', path='/ws')

        ```
        """
        parsed = parse_connection_string(connection_string)

        return ConnectionParams(**parsed)


def parse_connection_string(connection_string: str) -> dict:
    """
    Parse a connection string into a dictionary of connection parameters.

    ```pycon
    >>> parse_connection_string("ws://user:pass@host/ws")
    {'host': 'host',
        'port': 80,
        'protocol': 'ws',
        'username': 'user',
        'password': 'pass',
        'path': '/ws'}

    >>> parse_connection_string("wss://user:pass@host/ws")
    {'host': 'host',
        'port': 443,
        'protocol': 'wss',
        'username': 'user',
        'password': 'pass',
        'path': '/ws'}

    ```
    """

    parsed = urlparse(connection_string)

    default_port = 80 if parsed.scheme == "ws" else 443

    return {
        "host": parsed.hostname,
        "port": parsed.port or default_port,
        "protocol": parsed.scheme,
        "username": parsed.username,
        "password": parsed.password,
        "path": parsed.path,
    }



import asyncio
import json
import logging

from abc import abstractmethod
from enum import StrEnum, auto
from dataclasses import dataclass, field
from uuid import uuid4

from .exceptions import ErrorMessageException

logger = logging.getLogger(__name__)


class MessageType(StrEnum):
    incoming = auto()
    outgoing = auto()
    error = auto()
    response = auto()
    unknown = auto()


@dataclass
class ReceivedMessage:
    message_type: MessageType
    correlation_id: str
    data: dict


@dataclass
class BaseConnection:
    incoming_queue: asyncio.Queue = field(default_factory=asyncio.Queue)
    outgoing_queue: asyncio.Queue = field(default_factory=asyncio.Queue)
    error_queue: asyncio.Queue = field(default_factory=asyncio.Queue)
    result_queues: dict[str, asyncio.Queue] = field(default_factory=dict)

    echo: bool = False
    silent: bool = False

    @abstractmethod
    async def run(self):
        pass

    @abstractmethod
    async def send(
        self, topic_end: str, data: dict, correlation_data: str | None = None
    ):
        pass

    async def handle(self, message: ReceivedMessage):
        is_error = message.message_type == MessageType.error

        if self.echo:
            logger.info(
                "%s %s %s",
                message.type,
                message.correlation_id,
                json.dumps(message.data, indent=2),
            )

        if is_error and not self.silent:
            logger.error("Got exception %s", json.dumps(message.data))

        if message.correlation_id in self.result_queues:

            await self.result_queues[message.correlation_id].put(
                (is_error, message.data)
            )

        if message.message_type == MessageType.incoming:
            await self.incoming_queue.put(message.data)
        elif message.message_type == MessageType.outgoing:
            await self.outgoing_queue.put(message.data)
        elif message.message_type == MessageType.error:
            await self.error_queue.put(message.data)

    async def send_with_reply(self, topic_end: str, data: dict):
        correlation_id = str(uuid4())

        self.result_queues[correlation_id] = asyncio.Queue()

        await self.send(topic_end, data, correlation_data=correlation_id)

        async with asyncio.timeout(1):
            error, result = await self.result_queues[correlation_id].get()
        del self.result_queues[correlation_id]

        if error:
            raise ErrorMessageException(result)

        return result



import pytest
from unittest.mock import AsyncMock

from . import Almabtrieb


def test_unknown_connection_string():
    with pytest.raises(NotImplementedError):
        Almabtrieb.from_connection_string("unknown://user:pass@host/ws")


async def test_create_actor():
    almabtrieb = Almabtrieb(connection=AsyncMock())

    await almabtrieb.create_actor("http://host.example")

    almabtrieb.connection.send_with_reply.assert_called_once()



import pytest

from almabtrieb.util import parse_connection_string

from .base import MqttBaseConnection
from .test_mqtt import is_mqtt_installed


@pytest.mark.skip(
    reason="Testing connectivity only possible when MQTT broker is available"
)
async def test_base_mqtt():
    parsed = parse_connection_string("mqtt://almabtrieb:password@localhost:11883")

    connection = MqttBaseConnection(**parsed)

    print(connection)

    async with connection.run(topics_to_subscribe=[]):
        assert connection.connected


@pytest.mark.skipif(not is_mqtt_installed(), reason="aiomqtt not installed")
async def test_mqtts():
    parsed = parse_connection_string("mqtts://test.mosquitto.org:8883")

    connection = MqttBaseConnection(**parsed)

    print(connection)

    async with connection.run(topics_to_subscribe=[]):
        assert connection.connected
